module Utils = struct

let string_split_on sep str =
  let rec aux str ans =
    if str = "" then ans else
    try
      let first_space = String.index str sep in
      if first_space = 0 then
        aux (String.sub str 1 (String.length str - 1)) ans
      else
        aux
          (String.sub str (first_space + 1)(String.length str - 1 - first_space))
          ((String.sub str 0 (first_space)) :: ans)
    with Not_found -> List.rev (str :: ans)
  in aux str []
;;

(* left fold over every line of a file *)
let foldl_file f x file =
  let ic = Pervasives.open_in file
  and acc = ref x in
  (try while true do acc := f (!acc) (input_line ic) done
  with End_of_file -> close_in ic);
  !acc
;;

(* timeout in seconds *)
type timeout_t = Timeout of float;;

let parallelise f x n_processes (Timeout timeout) =
  (* starting times and PIDs of running threads *)
  let thrs = ref [] in
  let state = ref x in

  let add_thread f = match f (!state) with
    | None -> ()
    | Some (pid, state') ->
        thrs := (Unix.gettimeofday (), pid) :: !thrs;
        state := state' in

  (* check whether thread has exceeded its given time *)
  let exceeded (start_t, _) = start_t +. timeout < Unix.gettimeofday() in

  (* create initial thread pool, and whenever a thread is done, start new one *)
  for i = 0 to n_processes-1 do add_thread f done;
  while !thrs <> [] do
    (* HOL Light's parser cannot parse "Unix.WNOHANG", so we cast its value 0
       to the wait_flag type. *)
    let wnohang = ((Obj.magic 0) : Unix.wait_flag) in
    match fst (Unix.waitpid [wnohang] (-1)) with
      0 -> (* no process has recently terminated *)
        (* processes that have exceeded their time limit *)
        let late = List.filter exceeded !thrs in
        if late = [] then Unix.sleep 1 else
        (* kill those good-for-nothings! *)
        List.iter (fun (_, pid) -> try Unix.kill pid 9 with _ -> ()) late;
    | pid -> (* this process just finished *)
        (* remove finished thread from thread list *)
        thrs := List.filter (fun (_, p) -> p <> pid) !thrs;
        add_thread f;
  done;
  !state
;;

end
