module Metis_reconstruct = struct

let hol_of_var tymap v =
  try let v' = int_of_string v in
      Meson.hol_of_var v'
  with _ ->
  (match Name.Map.peek tymap v with
    Some ty -> mk_var (v, ty)
  | None -> failwith ("Metis_reconstruct.hol_of_var: " ^ v))

let hol_of_const c =
  try Meson.hol_of_const (int_of_string c)
  with _ -> failwith ("Metis_reconstruct.hol_of_const: " ^ c)

let rec hol_of_term tymap = function
    Term.Var v -> hol_of_var tymap v
  | Term.Fn (f, args) ->
      list_mk_comb(hol_of_const f, map (hol_of_term tymap) args)

let hol_of_atom tymap (p,args) =
  let args' = map (hol_of_term tymap) args in
  if p = "=" then match args' with
      eq_l :: eq_r :: [] -> mk_eq (eq_l, eq_r)
    | _ -> failwith "Metis_reconstruct.hol_of_atom: bad equality"
  else list_mk_comb(hol_of_const p,args')

let hol_of_literal tymap (pol,args) =
  let atom = hol_of_atom tymap args in
  match pol with
    false -> mk_neg atom
  | true  -> atom

let string_of_metis_subst subst =
  let print_single (name, term) = name ^ " -> " ^ Term.toString term in
  String.concat ", " (map print_single subst)

let string_of_hol_subst subst =
  let print_single (t, v) = string_of_term t ^ " / " ^ string_of_term v in
  String.concat ", " (map print_single subst)

let hol_of_subst tymap subst =
  map (fun (name, term) -> hol_of_term tymap term, hol_of_var tymap name) subst

let empty_tymap = Name.Map.newMap ()

(* update type map with all variables from substitution that are not covered
   by the type map yet *)
let update_tymap tymap subst =
  let f (v, term) acc =
    try let _ = hol_of_var acc v in acc
    with _ -> Name.Map.insert acc (v, type_of (hol_of_term tymap term)) in
  List.fold_right f subst tymap

let string_of_int_list = String.concat "; " o List.map string_of_int


(* return polarity and atom of literal *)
let atom_of_literal lit = match is_neg lit with
  true  -> (false, dest_neg lit)
| false -> (true ,          lit)

let literal_of_atom (pol, atom) = if pol then atom else mk_neg atom

let rec follow_metis_atom_path tm = function
    [] -> (tm, "")
  | i :: is ->
    let f,args = strip_comb tm in
    let arity = length args in
    if i < arity then
      let (tm', path') = follow_metis_atom_path (List.nth args i) is in
      (tm', String.make (arity - i - 1) 'l' ^ "r" ^ path')
    else failwith "follow_metis_atom_path"

(* find literal subterm at a Metis path, and return it along with
   equivalent HOL Light path *)
let follow_metis_lit_path lit path =
  (* Metis returns paths that indicate the position of a term inside an
     *atom*, even if the atom is negated, thus a literal. *)
  let (pol, atom) = atom_of_literal lit in
  let (s, path) = follow_metis_atom_path atom path in
  (s, if pol then path else "r" ^ path)


(* retrieve axiom that proves the disjunction of the given literals *)
let match_axiom axioms lits =
  let set = setify lits in
  let clause = list_mk_disj set in
  (* we could canonicalise just once at the beginning for all axioms *)
  let canonicalise = CONV_RULE DISJ_CANON_CONV in
  let axioms' = map canonicalise axioms in
  try find (fun thm -> concl thm = clause) axioms'
  with _ -> failwith "match_axiom"

let EXCLUDED_MIDDLE = TAUT `!a. a \/ ~a`

(* move a literal in the proof of a disjunction to the first position
   may not preserve the order of the other literals *)
let FRONT lit thm =
  let conc = concl thm in
  let disj = disjuncts (concl thm) in
  let rest = match partition (fun l -> l = lit) disj with
      ([], _) -> failwith "FRONT: literal not in disjunction"
    | (_ , r) -> r in
  let disj' = lit :: rest in
  let conc' = list_mk_disj disj' in
  let eq = DISJ_ACI_RULE (mk_eq (conc, conc')) in
  (PURE_ONCE_REWRITE_RULE [eq] thm, rest)

let RESOLVE_1  = TAUT `!a. a ==> ~a ==> F`
let RESOLVE_2L = TAUT `!a b. a \/ b ==> ~a ==> b`
let RESOLVE_2R = TAUT `!a c. a ==> ~a \/ c ==> c`
let RESOLVE_3  = TAUT `!a b c. a \/ b ==> ~a \/ c ==> b \/ c`

(* resolve two clauses, where atom has to appear at the first position of
   both clauses: positive in the first and negative in the second clause *)
let RESOLVE_N atom = function
  ([], []) -> SPEC atom RESOLVE_1
| (r1, []) -> SPECL [atom; list_mk_disj r1] RESOLVE_2L
| ([], r2) -> SPECL [atom; list_mk_disj r2] RESOLVE_2R
| (r1, r2) -> SPECL [atom; list_mk_disj r1; list_mk_disj r2] RESOLVE_3

(* resolve two clauses th1 and th2, where atom appears somewhere
   positive in th1 and negative in th2 *)
let RESOLVE atom th1 th2 =
  (*print_endline ("Atom: " ^ string_of_term atom);
  print_endline ("th1 : " ^ string_of_term (concl th1));
  print_endline ("th2 : " ^ string_of_term (concl th2));*)
  try let (th1', r1) = FRONT atom th1
  and (th2', r2) = FRONT (mk_neg atom) th2 in
  let res = RESOLVE_N atom (r1, r2) in
  MP (MP res th1') th2'
  with _ -> failwith "resolve"

let IMPL_NOT_L = TAUT `!a b. ~a ==> b <=>  a \/ b`;;
let IMPL_NOT_R = TAUT `!a b.  a ==> b <=> ~a \/ b`;;

(* given A,  tm |- C, prove A |- ~tm \/ C or
   given A, ~tm |- C, prove A |-  tm \/ C *)
let DISCH_DISJ tm th =
  let impl = DISCH tm th
  and (tm', IMPL_NOT) =
    try dest_neg tm, IMPL_NOT_L
    with _ ->    tm, IMPL_NOT_R in
  let eq = SPECL [tm'; concl th] IMPL_NOT in
  PURE_ONCE_REWRITE_RULE [eq] impl

(* given A, tm1, .., tmn |- th, prove A |- ~tm1 \/ .. \/ ~tmn \/ th *)
let DISCH_DISJS tms th = List.fold_right DISCH_DISJ tms th

let rec hol_of_proof axioms tymap th_concl =
  let hol_term  = hol_of_term tymap
  and hol_atom  = hol_of_atom tymap
  and hol_lit   = hol_of_literal tymap
  and hol_proof = hol_of_proof axioms tymap in

  (*print_string ("Thm: " ^ Thm.toString th_concl);
  print_newline ();*)

  let result =
  match Proof.thmToInference th_concl with
    Proof.Axiom litset ->
      let lits = Literal.Set.toList litset in
      match_axiom axioms (map hol_lit lits)
  | Proof.Assume atom -> SPEC (hol_atom atom) EXCLUDED_MIDDLE
  | Proof.Subst (subst, th) ->
      let sl = Substitute.toList subst in
      let tymap' = update_tymap tymap sl in
      let subst' = hol_of_subst tymap' sl in
      let proof' = (hol_of_proof axioms tymap' th) in
      (*print_endline ("Subst (Metis): " ^ string_of_metis_subst sl);
      print_endline ("Subst (HOL): " ^ string_of_hol_subst subst');
      print_string "Proof: ";
      print_thm proof';
      print_newline ();*)
      INST subst' proof'
  | Proof.Resolve (atom, th1, th2) ->
      RESOLVE (hol_atom atom) (hol_proof th1) (hol_proof th2)
  | Proof.Refl term -> REFL (hol_term term)
  | Proof.Equality (lit, path, term) ->
      let lit' = hol_lit lit
      and t = hol_term term in

      let (s, path') = follow_metis_lit_path lit' path in

      (*print_string "Literal: "; print_term lit';
      print_newline ();
      print_string "Path (Metis): "; print_string (string_of_int_list path);
      print_newline ();
      print_string "Path (HOL): "; print_string path';
      print_newline ();
      print_string "Term (t, Metis): "; print_string (Term.toString term);
      print_newline ();
      print_string "Term (t, HOL): "; print_term t;
      print_newline ();
      print_string "Term (s, from fmlp): "; print_term s;
      print_newline ();
      print_string "Term (s, from fp): "; print_term (follow_path path' lit');
      print_newline ();*)

      let eq = mk_eq (s,t) in
      let conv = PATH_CONV path' (PURE_ONCE_REWRITE_CONV [ASSUME eq]) in
      let converted = CONV_RULE conv (ASSUME lit') in
      (*print_thm converted;
      print_newline ();*)
      try DISCH_DISJS [eq; lit'] converted
      with _ -> failwith "equality"

  in
    (*print_endline ("Desired: " ^ Thm.toString th_concl);
    print_endline ("Result : " ^ string_of_term (concl result));*)
    result

end
