(* Functions copied from meson.ml, usable also in Leancop and Metis. *)

module Meson =
struct

let offinc = 10000

(* originally in fol_of_hol_clauses *)
let mk_negated (p,a) = -p,a

(* ------------------------------------------------------------------------- *)
(* Shadow syntax for FOL terms in NNF. Functions and predicates have         *)
(* numeric codes, and negation is done by negating the predicate code.       *)
(* ------------------------------------------------------------------------- *)

type fol_term = Fvar of int
              | Fnapp of int * fol_term list;;

type fol_atom = int * fol_term list;;

type fol_form = Atom of fol_atom
               | Conj of fol_form * fol_form
               | Disj of fol_form * fol_form
               | Forallq of int * fol_form;;

(* ----------------------------------------------------------------------- *)
(* Translate a term (in NNF) into the shadow syntax.                       *)
(* ----------------------------------------------------------------------- *)

let reset_vars,fol_of_var,hol_of_var =
  let vstore = ref []
  and gstore = ref []
  and vcounter = ref 0 in
  let inc_vcounter() =
    let n = !vcounter in
    let m = n + 1 in
    if m >= offinc then failwith "inc_vcounter: too many variables" else
    (vcounter := m; n) in
  let reset_vars() = vstore := []; gstore := []; vcounter := 0 in
  let fol_of_var v =
    let currentvars = !vstore in
    try assoc v currentvars with Failure _ ->
    let n = inc_vcounter() in
    vstore := (v,n)::currentvars; n in
  let hol_of_var v =
    try rev_assoc v (!vstore)
    with Failure _ -> rev_assoc v (!gstore) in
  let hol_of_bumped_var v =
    try hol_of_var v with Failure _ ->
    let v' = v mod offinc in
    let hv' = hol_of_var v' in
    let gv = genvar(type_of hv') in
    gstore := (gv,v)::(!gstore); gv in
  reset_vars,fol_of_var,hol_of_bumped_var

let reset_consts,fol_of_const,hol_of_const =
  let false_tm = `F` in
  let cstore = ref ([]:(term * int)list)
  and ccounter = ref 2 in
  let reset_consts() = cstore := [false_tm,1]; ccounter := 2 in
  let fol_of_const c =
    let currentconsts = !cstore in
    try assoc c currentconsts with Failure _ ->
    let n = !ccounter in
    ccounter := n + 1; cstore := (c,n)::currentconsts; n in
  let hol_of_const c = rev_assoc c (!cstore) in
  reset_consts,fol_of_const,hol_of_const

let rec fol_of_term env consts tm =
  if is_var tm & not (mem tm consts) then
    Fvar(fol_of_var tm)
  else
    let f,args = strip_comb tm in
    if mem f env then failwith "fol_of_term: higher order" else
    let ff = fol_of_const f in
    Fnapp(ff,map (fol_of_term env consts) args)

let fol_of_atom env consts tm =
  let f,args = strip_comb tm in
  if mem f env then failwith "fol_of_atom: higher order" else
  let ff = fol_of_const f in
  ff,map (fol_of_term env consts) args

let fol_of_literal env consts tm =
  try let tm' = dest_neg tm in
      let p,a = fol_of_atom env consts tm' in
      -p,a
  with Failure _ -> fol_of_atom env consts tm

let rec fol_of_form env consts tm =
  try let v,bod = dest_forall tm in
      let fv = fol_of_var v in
      let fbod = fol_of_form (v::env) (subtract consts [v]) bod in
      Forallq(fv,fbod)
  with Failure _ -> try
      let l,r = dest_conj tm in
      let fl = fol_of_form env consts l
      and fr = fol_of_form env consts r in
      Conj(fl,fr)
  with Failure _ -> try
      let l,r = dest_disj tm in
      let fl = fol_of_form env consts l
      and fr = fol_of_form env consts r in
      Disj(fl,fr)
  with Failure _ ->
      Atom(fol_of_literal env consts tm)

(* ----------------------------------------------------------------------- *)
(* Further translation functions for HOL formulas.                         *)
(* ----------------------------------------------------------------------- *)

let rec hol_of_term tm =
  match tm with
    Fvar v -> hol_of_var v
  | Fnapp(f,args) -> list_mk_comb(hol_of_const f,map hol_of_term args)

let hol_of_atom (p,args) =
  list_mk_comb(hol_of_const p,map hol_of_term args)

let hol_of_literal (p,args) =
  if p < 0 then mk_neg(hol_of_atom(-p,args))
  else hol_of_atom (p,args)

(* ----------------------------------------------------------------------- *)
(* Versions of shadow syntax operations with variable bumping.             *)
(* ----------------------------------------------------------------------- *)

let rec fol_free_in v tm =
  match tm with
    Fvar x -> x = v
  | Fnapp(_,lis) -> exists (fol_free_in v) lis

let rec fol_subst theta tm =
  match tm with
    Fvar v -> rev_assocd v theta tm
  | Fnapp(f,args) ->
        let args' = qmap (fol_subst theta) args in
        if args' == args then tm else Fnapp(f,args')

let fol_inst theta ((p,args) as at:fol_atom) =
  let args' = qmap (fol_subst theta) args in
  if args' == args then at else p,args'

let rec fol_subst_bump offset theta tm =
  match tm with
    Fvar v -> if v < offinc then
               let v' = v + offset in
               rev_assocd v' theta (Fvar(v'))
             else
               rev_assocd v theta tm
  | Fnapp(f,args) ->
        let args' = qmap (fol_subst_bump offset theta) args in
        if args' == args then tm else Fnapp(f,args')

let fol_inst_bump offset theta ((p,args) as at:fol_atom) =
  let args' = qmap (fol_subst_bump offset theta) args in
  if args' == args then at else p,args'

(* ----------------------------------------------------------------------- *)
(* Main unification function, maintaining a "graph" instantiation.         *)
(* We implicitly apply an offset to variables in the second term, so this  *)
(* is not symmetric between the arguments.                                 *)
(* ----------------------------------------------------------------------- *)

let rec istriv env x t =
  match t with
    Fvar y -> y = x or
              (try let t' = rev_assoc y env in istriv env x t'
               with Failure "find" -> false)
  | Fnapp(f,args) -> exists (istriv env x) args & failwith "cyclic"

let rec fol_unify offset tm1 tm2 sofar =
  match tm1,tm2 with
    Fnapp(f,fargs),Fnapp(g,gargs) ->
        if f <> g then failwith "" else
        itlist2 (fol_unify offset) fargs gargs sofar
  | _,Fvar(x) ->
       (let x' = x + offset in
        try let tm2' = rev_assoc x' sofar in
            fol_unify 0 tm1 tm2' sofar
        with Failure "find" ->
            if istriv sofar x' tm1 then sofar
            else (tm1,x')::sofar)
  | Fvar(x),_ ->
       (try let tm1' = rev_assoc x sofar in
            fol_unify offset tm1' tm2 sofar
        with Failure "find" ->
            let tm2' = fol_subst_bump offset [] tm2 in
            if istriv sofar x tm2' then sofar
            else (tm2',x)::sofar)

(* ----------------------------------------------------------------------- *)
(* Test for equality under the pending instantiations.                     *)
(* ----------------------------------------------------------------------- *)

let rec fol_eq insts tm1 tm2 =
  tm1 == tm2 or
  match tm1,tm2 with
    Fnapp(f,fargs),Fnapp(g,gargs) ->
        f = g & forall2 (fol_eq insts) fargs gargs
  | _,Fvar(x) ->
       (try let tm2' = rev_assoc x insts in
            fol_eq insts tm1 tm2'
        with Failure "find" ->
        try istriv insts x tm1 with Failure _ -> false)
  | Fvar(x),_ ->
       (try let tm1' = rev_assoc x insts in
            fol_eq insts tm1' tm2
        with Failure "find" ->
        try istriv insts x tm2 with Failure _ -> false)

let fol_atom_eq insts (p1,args1) (p2,args2) =
  p1 = p2 & forall2 (fol_eq insts) args1 args2


(* ----------------------------------------------------------------------- *)
(* Apply a multi-level "graph" instantiation.                              *)
(* ----------------------------------------------------------------------- *)

let rec fol_subst_partial insts tm =
  match tm with
    Fvar(v) -> (try let t = rev_assoc v insts in
                    fol_subst_partial insts t
                with Failure "find" -> tm)
  | Fnapp(f,args) -> Fnapp(f,map (fol_subst_partial insts) args)



(* ----------------------------------------------------------------------- *)
(* Creation of tagged contrapositives from a HOL clause.                   *)
(* This includes any possible support clauses (1 = falsity).               *)
(* The rules are partitioned into association lists.                       *)
(* ----------------------------------------------------------------------- *)

let fol_of_hol_clauses =
  let eqt (a1,(b1,c1)) (a2, (b2,c2)) =
   ((a1 = a2) & (b1 = b2) & (equals_thm c1 c2)) in
  let rec mk_contraposes n th used unused sofar =
    match unused with
      [] -> sofar
    | h::t -> let nw = (map mk_negated (used @ t),h),(n,th) in
              mk_contraposes (n + 1) th (used@[h]) t (nw::sofar) in
  let fol_of_hol_clause th =
    let lconsts = freesl (hyp th) in
    let tm = concl th in
    let hlits = disjuncts tm in
    let flits = map (fol_of_literal [] lconsts) hlits in
    let basics = mk_contraposes 0 th [] flits [] in
    if forall (fun (p,_) -> p < 0) flits then
      ((map mk_negated flits,(1,[])),(-1,th))::basics
    else basics in
  fun thms ->
    let rawrules = itlist (union' eqt o fol_of_hol_clause) thms [] in
    let prs = setify (map (fst o snd o fst) rawrules) in
    let prules =
      map (fun t -> t,filter ((=) t o fst o snd o fst) rawrules) prs in
    let srules = sort (fun (p,_) (q,_) -> abs(p) <= abs(q)) prules in
    srules


(* ----------------------------------------------------------------------- *)
(* Optimize set of clauses; changing literal order complicates HOL stuff.  *)
(* ----------------------------------------------------------------------- *)

let optimize_rules =
  let optimize_clause_order cls =
    sort (fun ((l1,_),_) ((l2,_),_) -> length l1 <= length l2) cls in
  map (fun (a,b) -> a,optimize_clause_order b)


(* ----------------------------------------------------------------------- *)
(* Create a HOL contrapositive on demand, with a cache.                    *)
(* ----------------------------------------------------------------------- *)

let clear_contrapos_cache,make_hol_contrapos =
  let DISJ_AC = AC DISJ_ACI
  and imp_CONV = REWR_CONV(TAUT `a \/ b <=> ~b ==> a`)
  and push_CONV =
    GEN_REWRITE_CONV TOP_SWEEP_CONV
     [TAUT `~(a \/ b) <=> ~a /\ ~b`; TAUT `~(~a) <=> a`]
  and pull_CONV = GEN_REWRITE_CONV DEPTH_CONV
     [TAUT `~a \/ ~b <=> ~(a /\ b)`]
  and imf_CONV = REWR_CONV(TAUT `~p <=> p ==> F`) in
  let memory = ref [] in
  let clear_contrapos_cache() = memory := [] in
  let make_hol_contrapos (n,th) =
    let tm = concl th in
    let key = (n,tm) in
    try assoc key (!memory) with Failure _ ->
    if n < 0 then
      CONV_RULE (pull_CONV THENC imf_CONV) th
    else
      let djs = disjuncts tm in
      let acth =
        if n = 0 then th else
        let ldjs,rdjs = chop_list n djs in
        let ndjs = (hd rdjs)::(ldjs@(tl rdjs)) in
        EQ_MP (DISJ_AC(mk_eq(tm,list_mk_disj ndjs))) th in
      let fth =
        if length djs = 1 then acth
        else CONV_RULE (imp_CONV THENC push_CONV) acth in
      (memory := (key,fth)::(!memory); fth) in
  clear_contrapos_cache,make_hol_contrapos


  (* from meson_to_hol *)
  let finish_RULE =
    GEN_REWRITE_RULE I
     [TAUT `(~p ==> p) <=> p`; TAUT `(p ==> ~p) <=> ~p`]


(* ----------------------------------------------------------------------- *)
(* Create equality axioms for all the function and predicate symbols in    *)
(* a HOL term. Not very efficient (but then neither is throwing them into  *)
(* automated proof search!)                                                *)
(* ----------------------------------------------------------------------- *)

let create_equality_axioms =
  let eq_thms = (CONJUNCTS o prove)
   (`(x:A = x) /\
     (~(x:A = y) \/ ~(x = z) \/ (y = z))`,
    REWRITE_TAC[] THEN ASM_CASES_TAC `x:A = y` THEN
    ASM_REWRITE_TAC[] THEN CONV_TAC TAUT) in
  let imp_elim_CONV = REWR_CONV
    (TAUT `(a ==> b) <=> ~a \/ b`) in
  let eq_elim_RULE =
    MATCH_MP(TAUT `(a <=> b) ==> b \/ ~a`) in
  let veq_tm = rator(rator(concl(hd eq_thms))) in
  let create_equivalence_axioms (eq,_) =
    let tyins = type_match (type_of veq_tm) (type_of eq) [] in
    map (INST_TYPE tyins) eq_thms in
  let rec tm_consts tm acc =
    let fn,args = strip_comb tm in
    if args = [] then acc
    else itlist tm_consts args (insert (fn,length args) acc) in
  let rec fm_consts tm ((preds,funs) as acc) =
    try fm_consts(snd(dest_forall tm)) acc with Failure _ ->
    try fm_consts(snd(dest_exists tm)) acc with Failure _ ->
    try let l,r = dest_conj tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
        let l,r = dest_disj tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
        let l,r = dest_imp tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
         fm_consts (dest_neg tm) acc with Failure _ ->
    try let l,r = dest_eq tm in
        if type_of l = bool_ty
        then fm_consts r (fm_consts l acc)
        else failwith "atomic equality"
    with Failure _ ->
    let pred,args = strip_comb tm in
    if args = [] then acc else
    insert (pred,length args) preds,itlist tm_consts args funs in
  let create_congruence_axiom pflag (tm,len) =
    let atys,rty = splitlist (fun ty -> let op,l = dest_type ty in
                                        if op = "fun" then hd l,hd(tl l)
                                        else fail())
                             (type_of tm) in
    let ctys = fst(chop_list len atys) in
    let largs = map genvar ctys
    and rargs = map genvar ctys in
    let th1 = rev_itlist (C (curry MK_COMB)) (map (ASSUME o mk_eq)
        (zip largs rargs)) (REFL tm) in
    let th2 = if pflag then eq_elim_RULE th1 else th1 in
    itlist (fun e th -> CONV_RULE imp_elim_CONV (DISCH e th)) (hyp th2) th2 in
  fun tms -> let preds,funs = itlist fm_consts tms ([],[]) in
             let eqs0,noneqs = partition
                (fun (t,_) -> is_const t & fst(dest_const t) = "=") preds in
             if eqs0 = [] then [] else
             let pcongs = map (create_congruence_axiom true) noneqs
             and fcongs = map (create_congruence_axiom false) funs in
             let preds1,_ =
               itlist fm_consts (map concl (pcongs @ fcongs)) ([],[]) in
             let eqs1 = filter
               (fun (t,_) -> is_const t & fst(dest_const t) = "=") preds1 in
             let eqs = union eqs0 eqs1 in
             let equivs =
               itlist (union' equals_thm o create_equivalence_axioms)
                      eqs [] in
             equivs@pcongs@fcongs


(* ----------------------------------------------------------------------- *)
(* Push duplicated copies of poly theorems to match existing assumptions.  *)
(* ----------------------------------------------------------------------- *)

let POLY_ASSUME_TAC =
  let rec uniq' eq =
    fun l ->
      match l with
        x::(y::_ as t) -> let t' = uniq' eq t in
                            if eq x y then t' else
                            if t'==t then l else x::t'
      | _ -> l in
  let setify' le eq s = uniq' eq (sort le s) in
  let rec grab_constants tm acc =
    if is_forall tm or is_exists tm then grab_constants (body(rand tm)) acc
    else if is_iff tm or is_imp tm or is_conj tm or is_disj tm then
      grab_constants (rand tm) (grab_constants (lhand tm) acc)
    else if is_neg tm then grab_constants (rand tm) acc
    else union (find_terms is_const tm) acc in
  let match_consts (tm1,tm2) =
    let s1,ty1 = dest_const tm1
    and s2,ty2 = dest_const tm2 in
    if s1 = s2 then type_match ty1 ty2 []
    else failwith "match_consts" in
  let polymorph mconsts th =
    let tvs = subtract (type_vars_in_term (concl th))
                       (unions (map type_vars_in_term (hyp th))) in
    if tvs = [] then [th] else
    let pconsts = grab_constants (concl th) [] in
    let tyins = mapfilter match_consts
      (allpairs (fun x y -> x,y) pconsts mconsts) in
    let ths' =
      setify' (fun th th' -> dest_thm th <= dest_thm th')
              equals_thm (mapfilter (C INST_TYPE th) tyins) in
    if ths' = [] then
      (warn true "No useful-looking instantiations of lemma"; [th])
    else ths' in
  let rec polymorph_all mconsts ths acc =
    if ths = [] then acc else
    let ths' = polymorph mconsts (hd ths) in
    let mconsts' = itlist grab_constants (map concl ths') mconsts in
    polymorph_all mconsts' (tl ths) (union' equals_thm ths' acc) in
  fun ths (asl,w as gl) ->
    let mconsts = itlist (grab_constants o concl o snd) asl [] in
    let ths' = polymorph_all mconsts ths [] in
    MAP_EVERY ASSUME_TAC ths' gl


let QUANT_BOOL_CONV =
  PURE_REWRITE_CONV[FORALL_BOOL_THM; EXISTS_BOOL_THM; COND_CLAUSES;
                    NOT_CLAUSES; IMP_CLAUSES; AND_CLAUSES; OR_CLAUSES;
                    EQ_CLAUSES; FORALL_SIMP; EXISTS_SIMP]

let CONJUNCTS_THEN' ttac cth =
  ttac(CONJUNCT1 cth) THEN ttac(CONJUNCT2 cth)


let rec SPLIT_TAC n g =
  ((FIRST_X_ASSUM(CONJUNCTS_THEN' ASSUME_TAC) THEN SPLIT_TAC n) ORELSE
   (if n > 0 then FIRST_X_ASSUM DISJ_CASES_TAC THEN SPLIT_TAC (n - 1)
    else NO_TAC) ORELSE
   ALL_TAC) g

end
